/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine;

import java.util.*;
import org.xml.sax.*;

public class PacketEngine extends Thread {

    private PacketEngineCore core;
    private String iXMLSpec = null;
    private boolean die = false;

    /**
     * Default constructor
     */
    public PacketEngine() {
	PacketUtils.debug = 1;
	core = new PacketEngineCore();
    }

    /**
     * Constructor.
     * Outputs debug information via stdout.
     * @param debugLevel Level of debug output: 0 = none, 5 = everything, 1 = default
     */
    public PacketEngine(int debugLevel) {
	PacketUtils.debug = debugLevel;
	core = new PacketEngineCore();
    }

    /**
     * Constructor.
     * @param out Outputmedia for debug messages
     * @param err Outputmedia for error messages
     * @param debugLevel Level of debug output: 0 = none, 5 = everything, 1 = default
     */
    public PacketEngine(MessageOutput out, MessageOutput err, int debugLevel) {
	PacketUtils.debug = debugLevel;
	PacketUtils.out = out;
	PacketUtils.err = err;
	core = new PacketEngineCore();
    }

    // PacketEngine methods

    /**
     * Set the die flag 
     */
    public synchronized void die() {
	die = true;
	notifyAll();
    }

    /**
     * Set XML specs
     *
     * @param XMLSpec The packetset specification
     * @return true if all went fine, false otherwise
     */
    public synchronized boolean setXMLSpec(String XMLSpec) {
	if(iXMLSpec != null) {
	    try{
		wait();
	    }
	    catch(InterruptedException exp) {
	    }
	}

	iXMLSpec = XMLSpec;
	notifyAll();
	return true;
    }
    
    /**
     * Get XML specs
     *
     * @return XMLSpec The packetset specification
     */
    private synchronized String waitForXMLSpec() {
	if(iXMLSpec == null) {
	    try{
		wait();
	    }
	    catch(InterruptedException exp) {
	    }
	}

	String spec = iXMLSpec;
	iXMLSpec = null;
	notifyAll();
	return spec;
    }
    
    /**
     * PacketEngine mainloop
     */
    public synchronized void run() {
	while(!halt()) {
	    String spec = waitForXMLSpec();
	    if(spec == null)
		continue;

	    core.start(spec);
	}
    }

    /**
     * Get the tag for the NetworkInterface.
     * @return The tag for the networkInterface or
     *         null if interface is not set.
     */
    public synchronized String getNetworkInterface() {
	String retval = core.getNetworkInterface();
	notifyAll();
	return(retval);
    }

    /**
     * Set the NetworkInterface.
     * @param if The network interface tag
     * @return true if operation was successfull,
     *         false otherwise
     */
    public synchronized boolean setNetworkInterface(String tag, String params) {
	boolean retval = core.setNetworkInterface(tag, params); 
	notifyAll();
	return(retval);
    }
    
    // private PacketEngine methods
    
    /**
     * Get the die flag
     * @return false if core is supposed to continue, 
     *         true otherwise
     */
    private synchronized boolean halt() {
	return(die);
    }

    // Static PacketEngine methods

    /**
     * Returns an array of Strings identifying 
     * the possible network output methods.
     */
    public static String[] getNetworkInterfaces() {
	return(PacketEngineCore.getNetworkInterfaces());
    }

    /**
     * Returns a Vector of ProtocolDescriptors.
     */
    public static Vector getProtocolTags() {
	return(PacketEngineCore.getProtocolTags());
    }

    /**
     * Returns the root tag for the XMLSpec
     */
    public static String getDocumentTag() {
        return(PacketEngineCore.getDocumentTag());
    }

    /**
     * Returns the packet tag for the XMLSpec
     */
    public static String getPacketTag() {
        return(PacketEngineCore.getPacketTag());
    }

    /**
     * Returns the sendtime attribute tag for the packet tag
     */
    public static String getPacketStimeTag() {
        return(PacketEngineCore.getPacketStimeTag());
    }

    /**
     * Returns the header tag for the XMLSpec
     */
    public static String getHeaderTag() {
        return(PacketEngineCore.getHeaderTag());
    }

    /**
     * Returns the type attribute tag for the header tag
     */
    public static String getHeaderTypeTag() {
        return(PacketEngineCore.getHeaderTypeTag());
    }

    /**
     * Set the debugLevel
     */
    public static void setDebugLevel(int debugLevel) {
	PacketUtils.debug = debugLevel;
    }
}
