/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine;

import packetengine.header.*;
import packetsender.*;

import java.util.*;
import java.io.*;
import java.net.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;


public class PacketEngineCore extends DefaultHandler {
    
    /* private variables */

    /* XML stuff */
    private SAXParserFactory factory;
    private SAXParser parser;

    /* current packet and list of packets */
    private Packet packet;
    private LinkedList packetList;

    /* networkinterfaces */
    private NetworkInterface networkIF;
    private static String[] networkInterfaces; 

    /* protocoltypes */
    private String[] protocolTypes = null;

    /* configfile */
    private static final String CONFIG_FILE = "PacketEngine.cfg";


    /**
     * Constructor.
     */
    public PacketEngineCore() {
        // Use the default (non-validating) parser
	factory = SAXParserFactory.newInstance();
	parser = null;

	// Initialize variables
	packet = null;
	packetList = null;
	networkIF = null;
	
	// Read properties file
	readConfig();

	// Default network interface
	try {
	    setNetworkInterface((String) Class.forName(networkInterfaces[0]).getMethod("getIFTag", null).invoke(null, null), null);
	}
	catch(Exception exp) {
	    PacketUtils.errdebugln("FATAL: " + exp);
	}

	/* Print debug stuff */
	PacketUtils.debugln(1, "\n");
	PacketUtils.debugln(1, "******************************");
	PacketUtils.debugln(1, "* PacketEngine instantiated  *");
	PacketUtils.debugln(1, "******************************");
	PacketUtils.debugln(1, "\n");
    }

    private static void readConfig(){
	String[] tmp;
	Properties p = new Properties();
	try{
	    p.load(new FileInputStream(CONFIG_FILE));
	}
	catch(Exception e){
	    PacketUtils.errdebugln("Error while loading " + CONFIG_FILE + ": " + e);
	    return;
	}
	String if_line = p.getProperty("IFs");
	
	StringTokenizer tk = new StringTokenizer(if_line,";");
	tmp = new String[tk.countTokens()];
	int i = 0; 
	while(tk.hasMoreTokens()){
	    tmp[i] = tk.nextToken().trim();
	    i++;
	}
	networkInterfaces = tmp;
    }

    /**
     * Get the NetworkInterface.
     * @return The tag for the networkInterface or
     *         null if interface is not set.
     */
    public String getNetworkInterface() {
	if(networkIF != null) {
	    try {
		return((String) networkIF.getClass().getMethod("getIFTag", null).invoke(null, null));
	    }
	    catch(Exception exp) {
		exp.printStackTrace();
	    }
	}
	return(null);
    }

    /**
     * Set the NetworkInterface.
     * @param tag The network interface tag
     * @param params Parameters for the network interface
     * @return true if operation was successfull,
     *         false otherwise
     */
    public boolean setNetworkInterface(String tag, String params) {

	/* Deactivate current interface */
	if(networkIF != null)
	    networkIF = null;
	
	/* Create new interface */
	for(int i=0; i<networkInterfaces.length; i++) {
	    try {
		Class c = Class.forName(networkInterfaces[i]);
		if(c != null && tag != null && tag.trim().equals(((String) c.getMethod("getIFTag", null).invoke(null, null)).trim())) {
		    networkIF = (NetworkInterface) c.newInstance();
		}
	    }
	    catch(Exception exp) {
		PacketUtils.errdebugln("Network interface error: could not instantiate network interface!");
		return(false);
	    }
	}

	/* Activate new interface */
	if(networkIF != null) {
	    try {
		if(networkIF.setParams(params)) {
		    PacketUtils.debugln(1, "Network interface " + tag + " configured!");
		    return(true);
		}
	    }
	    catch(Exception exp) {
	    }
	    PacketUtils.errdebugln("Network interface error: Invalid parameters!");
	    networkIF = null;
	}

	/* else */
	PacketUtils.errdebugln("Network interface error: Unknown error!");	
	return(false);
    }

    /**
     * Start operating.
     *
     * @param XMLSpec The XML input data
     */
    public boolean start(String XMLSpec) {
	return(start(new InputSource(new StringReader(XMLSpec))));
    }
    
    /**
     * Start operating.
     *
     * This method calls other methods to complete following tasks:
     * <ul>
     * <li>Parse the XML input</li>
     * <li>Create packets and protocol headers</li>
     * <li>Build packets</li>
     * <li>Send out packets</li>
     * </ul>
     * 
     * @param is The XML input data
     */
    public boolean start(InputSource is) {

	/* Print debug stuff */
	PacketUtils.debugln(1, "\n");
	PacketUtils.debugln(1, "******************************");
	PacketUtils.debugln(1, "*    PacketEngine starting   *");
	PacketUtils.debugln(1, "******************************");
	PacketUtils.debugln(1, "\n");

	/* Parse input, create and build packets */
	packetList = new LinkedList();
	try {
	    PacketUtils.debugln(5, "PacketEngineCore.start() - Parsing input");
	    // Parse the input 
	    parser = factory.newSAXParser();
	    parser.parse(is, this);
	    PacketUtils.debugln(5, "PacketEngineCore.start() - Input successfully parsed");
	}
	catch(NumberFormatException e) {
	    PacketUtils.errdebugln("Invalid Packetset Specification: Parameter Problem");
	    PacketUtils.errdebugln("Aborting operation.");
	    //e.printStackTrace();
	    return(false);
	}
	catch(Throwable t) {
	    PacketUtils.errdebugln("Invalid Packetset Specification: Parse Error");
	    PacketUtils.errdebugln("Aborting operation.");
	    //t.printStackTrace();
	    return(false);
	}

	/* Activate interface */
	if(networkIF == null) {
	    PacketUtils.errdebugln("FATAL: No network interface!");
	    return(false);
	}
	try {
	    if(!networkIF.ifup()) {
		PacketUtils.errdebugln("Network interface error: interface not activated!");
		return(false);
	    }
	}
	catch(Exception exp) {
	    PacketUtils.errdebugln("Network interface error: " + exp);
	    return(false);
	}
	PacketUtils.debugln(1, "Network interface activated!");
	
	/* Send packets at specified time */
	PacketUtils.debugln(1, "\nStart of packetset\t" + Calendar.getInstance().getTime() + "\n");

	long start = System.currentTimeMillis();	
	try {
	    for(int i=1; true; i++) {
		Packet packet = (Packet) packetList.removeFirst();
		try {
		    Thread.sleep(Math.max(0, start + packet.getSTime() - System.currentTimeMillis()));
		}
		catch(InterruptedException e) {
		}
		if(send(packet)) {
		    Calendar cal = Calendar.getInstance();
		    PacketUtils.debugln(1, "Packet " + i + " sent\t" 
					+ PacketUtils.formatIntValue(cal.get(Calendar.HOUR_OF_DAY), 2) + ":"
					+ PacketUtils.formatIntValue(cal.get(Calendar.MINUTE), 2) + ":"
					+ PacketUtils.formatIntValue(cal.get(Calendar.SECOND), 2) + "."
					+ PacketUtils.formatIntValue(cal.get(Calendar.MILLISECOND), 3));
		    PacketUtils.dumpPacket(5, packet);
		}
		else
		    break;
		
	    }
	}
	catch(NoSuchElementException e) {
	}

	PacketUtils.debugln(1, "\nEnd of packetset\t" + Calendar.getInstance().getTime() + "\n");
	
	/* Deactivate current interface */
	try {
	    if(!networkIF.ifdown()) {
		PacketUtils.errdebugln("Network interface error: ifdown aborted!");
		return(false);
	    }
	}
	catch(Exception exp) {
	    PacketUtils.errdebugln("Network interface error: " + exp);
	    return(false);
	}
	PacketUtils.debugln(1, "Network interface deactivated!");

	return(true);
    }

    /**
     * Sends a packet.
     * @param packet A single packet
     * @return true if everything went fine,
     *         false otherwise
     */
    private boolean send(Packet packet) {

	/* Tss metodissa paketti lhetetn etipin */

	/* Tll hetkell vain debuggaustulostusta... */
	byte[] packetBuffer = packet.getBytes();
	
	if(networkIF == null) {
	    PacketUtils.errdebugln("Send operation aborted: No network interface specified!");
	    return(false);
	}
	if(packetBuffer != null && packetBuffer.length > 0) {
	    if(networkIF.sendData(packetBuffer))
		return(true);

	    PacketUtils.errdebugln("Send operation aborted: Network interface error!");
	    return(false);
	}

	PacketUtils.errdebugln("Send operation aborted: Discarding empty packet!");
	return(false);
    }

    /**
     * Returns an array of protocoltypetags such that
     * type of protocols[i] equals protocolTypes[i].
     * @return The typetags
     */
    private String[] getProtocolTypes() {
	if(protocolTypes != null)
	    return(protocolTypes);

	protocolTypes = new String[protocols.length];
	for(int i=0; i<protocols.length; i++) {
	    try {
		if(!protocols[i].startsWith(SEPARATOR))
		    protocolTypes[i] = (String) Class.forName(Header.PACKAGE_PREFIX + "." + protocols[i]).getMethod("getType", null).invoke(null, null);
		else
		    protocolTypes[i] = null;
	    }
	    catch(ClassNotFoundException exp) {
		PacketUtils.errdebugln(exp.toString());		
	    }
	    catch(Exception exp) {
		exp.printStackTrace();
		protocolTypes = null;
		break;
	    }
	}

	return(protocolTypes);
    }

    /* DefaultHandler methods */
    
    /**
     * SAXParser callback method
     */
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {

	PacketUtils.debug(5, "startElement >\tqName: " + qName + "\tAttr: ");
	for(int i=0; i<attributes.getLength(); i++)
	    PacketUtils.debug(5, attributes.getQName(i) + ": " + attributes.getValue(i));
	PacketUtils.debugln(5, "");

	String tag = qName.trim();

	if(Header.TAG_DOC.equals(tag)) {
	    PacketUtils.debugln(5, "XML Spec begins:");
	}
	   
	else if(Header.TAG_PACKET.equals(tag)) {
	    PacketUtils.debugln(5, "New Packet");
	    packet = new Packet();
	    
	    int stime = 0;
	    for(int i=0; i<attributes.getLength(); i++) {
		if(Header.TAG_PACKET_ATTR_STIME.equals(attributes.getQName(i)))
		    stime = Integer.parseInt(attributes.getValue(i));
	    }
	    packet.setSTime(stime);
	}

	else if(Header.TAG_HEADER.equals(tag)) {
	    String headerType = null;
	    Header h = null;
	    for(int i=0; i<attributes.getLength(); i++) {
		if(Header.TAG_HEADER_ATTR_TYPE.equals(attributes.getQName(i))) {
		    headerType = attributes.getValue(i).trim();
		    break;
		}
	    }
	    if(headerType == null)
		PacketUtils.debugln(5, "No type field found!");

	    else {
		PacketUtils.debugln(5, "New Header: " + headerType);

		String[] types = getProtocolTypes();
		try {
		    for(int i=0; i<types.length; i++) {
			if(types[i] != null && types[i].equals(headerType)) {
			    h = (Header) Class.forName(Header.PACKAGE_PREFIX + "." + protocols[i]).newInstance();
			    break;
			}
		    }
		}
		catch(ClassNotFoundException e) {
		    e.printStackTrace();
		    h = null;
		}
		catch(InstantiationException e) {
		    e.printStackTrace();
		    h = null;
		}
		catch(IllegalAccessException e) {
		    e.printStackTrace();
		    h = null;
		}
		catch(NullPointerException e) {
		    e.printStackTrace();
		    h = null;
		}
	    }
	    
	    if(h != null) {
		h.setHandlers(this, parser);
		try {
		    parser.getXMLReader().setContentHandler((DefaultHandler) h);
		    packet.add(h);
		}
		catch(Exception e) {
		    PacketUtils.errdebugln("FATAL:" + e.toString());
		    return;
		}
	    }
	}
    }

    /**
     * SAXParser callback method
     */
    public void endElement(String uri, String localName, String qName) throws SAXException {
	PacketUtils.debugln(5, "endElement >\tqName: " + qName);

	if(Header.TAG_PACKET.equals(qName)) {
	    packetList.add(packet);
	    packet.build();
	}
    }

    /**
     * Returns an array of Strings identifying 
     * the possible network output methods.
     * @return An array of Strings
     */
    public static String[] getNetworkInterfaces() {
	if (networkInterfaces == null)
	    readConfig();
	if (networkInterfaces == null)
	    return(new String[0]);

	String[] ids = new String[networkInterfaces.length];

	for(int i=0, j=0; i<networkInterfaces.length; i++) {
	    try {
		ids[j] = ((String) Class.forName(networkInterfaces[i]).getMethod("getIFTag", null).invoke(null, null)).trim();
	    }
	    catch(Exception exp) {
		PacketUtils.errdebugln("FATAL: " + exp);
		ids[j] = "";
	    }
	    j++;
	}

	return(ids);
    }

    // Registered Protocols
    private static final String SEPARATOR = "SEP_";
    private static final String[] protocols = { SEPARATOR + "IPv4",
	                                        "IPv4",
						"IPv4_OPT_EOL",
						"IPv4_OPT_NOOP",
						"IPv4_OPT_LSRR",
						"IPv4_OPT_SSRR",
						"IPv4_OPT_RR",
						"IPv4_OPT_SEC",
						"IPv4_OPT_SID",
						"IPv4_OPT_TS",
						SEPARATOR + "IPv6",
						"IPv6",
						"IPv6_HOP",
						"IPv6_OPT_PAD1",
						"IPv6_OPT_PADN",
						"IPv6_ROUTING",
						"IPv6_OPT_ROUTING_TYPE0",
						"IPv6_FRAGMENT",
						"IPv6_DESTINATION",
						"IPv6_OPT_DESTINATION_BINDING_UPDATE",
						"IPv6_OPT_DESTINATION_BINDING_ACK",
						"IPv6_OPT_DESTINATION_BINDING_REQ",
						"IPv6_OPT_DESTINATION_HOME_ADDRESS",
						"IPv6_OPT_SUB_BINDING_UNIQUE_ID",
						"IPv6_OPT_SUB_BINDING_ALT_COA",
						"IPv6_OPT_SUB_BINDING_AUTH",
						SEPARATOR + "ICMPv4",
						"ICMPv4",
						SEPARATOR + "TCP/UDP",
						"TCP",
						"UDP",
						SEPARATOR + "MIPv4",
						"MIPv4_RegReq",
						"MIPv4_RegRep",
						SEPARATOR + "Other",
						"Payload",
    };						

    /**
     * Returns a vector of valid XML tags for supported protocol headers.
     *
     * @return a Vector of <code>ProtocolDescriptor</code> objects
     */
    public static Vector getProtocolTags() {
	Vector v = new Vector();

	for(int i=0; i<protocols.length; i++) {
	    try {
		if(protocols[i].startsWith(SEPARATOR))
		    v.add(protocols[i].substring(SEPARATOR.length()));
		else {
		    Class c = Class.forName(Header.PACKAGE_PREFIX + "." + protocols[i]);
		    ProtocolDescriptor p = new ProtocolDescriptor(((String) c.getMethod("getType", null).invoke(null, null)).trim(),
								  ((String) c.getMethod("getName", null).invoke(null, null)).trim(),
								  (LinkedHashMap) c.getMethod("getTags", null).invoke(null, null));
		    v.add(p);
		}
	    }
	    catch(Exception exp) {
		PacketUtils.errdebugln("FATAL: " + exp);
	    }
	}
	
	return(v);
    }

    /**
     * Returns the root tag for the XML spec.
     * @return The tag
     */
    public static String getDocumentTag() {
	return(Header.TAG_DOC);
    }

    /**
     * Returns the packet delimiter tag for the XML spec.
     * @return The tag
     */
    public static String getPacketTag() {
	return(Header.TAG_PACKET);
    }

    /**
     * Returns the send time attribute tag for the XML spec.
     * @return The tag
     */
    public static String getPacketStimeTag() {
	return(Header.TAG_PACKET_ATTR_STIME);
    }

    /**
     * Returns the header delimiter tag for the XML spec.
     * @return The tag
     */    
    public static String getHeaderTag() {
	return(Header.TAG_HEADER);
    }

    /**
     * Returns the header type attribute tag for the XML spec.
     * @return The tag
     */
    public static String getHeaderTypeTag() {
	return(Header.TAG_HEADER_ATTR_TYPE);
    }
    
    // Rest is just for debugging, remove if this causes any problems
    
    public static void main(String[] args) {
	PacketEngineCore p = new PacketEngineCore();
	InputSource is;
	for(int i=0; i<args.length; i++) {
	    if("-d".equals(args[i]) || "--debug".equals(args[i]))
		PacketUtils.debug = 5;
	    else {
		try {
		    System.out.println("Current netIF:\t\t" + p.getNetworkInterface());
		    String[] tags = PacketEngineCore.getNetworkInterfaces();

		    p.setNetworkInterface(tags[0], null);

		    is = new InputSource(new FileReader(args[i]));
		    p.start(is);
		}
		catch(Exception e) {
		    e.printStackTrace();
		}
	    }
	}
    }
}
