/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine;

import packetengine.header.*;

import java.util.*;
import java.net.*;

public class PacketUtils {

    public static int debug = 5; /* 0 = none, 5 = all */ 
    public static boolean errdebug = true;
    public static MessageOutput out = null;
    public static MessageOutput err = null;

    /**
     * Output a string of debugdata.
     * @param level The level of importance of the message
     * @param s The debug message
     */
    public static void debug(int level, String s) {
	if(level <= debug) {
	    if(out != null)
		out.print(s);
	    else
		System.out.print(s);
	}
    }

    /**
     * Output a line of debugdata.
     * @param level The level of importance of the message
     * @param s The debug message
     */
    public static void debugln(int level, String s) {
	if(level <= debug) {
	    if(out != null)
		out.println(s);
	    else
		System.out.println(s);
	}
    }

    /**
     * Output an error message.
     * @param s The error message
     */
    public static void errdebugln(String s) {
	if(errdebug) {
	    if(err != null) {
		err.popup();
		err.println(s);
	    }
	    else
		System.err.println(s);
	}
    }

    /**
     * Convert a string to an byte.
     * If the string has a "0x" prefix, the value is 
     * interpreted in hexadecimal format.
     * @param strValue The string
     * @return The value
     */
    public static byte parseByte(String strValue) {
	int value = 0;

	try {
	    if(strValue.startsWith("0x")) {
		strValue = strValue.substring(2);
		value = Integer.parseInt(strValue, 16);
	    }
	    else
		value = Integer.parseInt(strValue);

	    if(value > 255 || value < 0)
		throw new NumberFormatException();
	}
	catch(NumberFormatException exp) {
	    errdebugln("Invalid number format: " + strValue);
	    throw exp;
	}

	return((byte) (value & 0xff));
    }

    /**
     * Convert a string to an int.
     * If the string has a "0x" prefix, the value is 
     * interpreted in hexadecimal format.
     * @param strValue The string
     * @return The value
     */
    public static int parseInt(String strValue) {
	long value = 0;

	try {
	    if(strValue.startsWith("0x")) {
		strValue = strValue.substring(2);
		value = Long.parseLong(strValue, 16);
	    }
	    else
		value = Long.parseLong(strValue);
	}
	catch(NumberFormatException exp) {
	    errdebugln("Invalid number format: " + strValue);
	    throw exp;
	}

	return((int) (value & 0xffffffff));
    }

    /**
     * Convert a string to an long.
     * If the string has a "0x" prefix, the value is 
     * interpreted in hexadecimal format.
     * @param strValue The string
     * @return The value
     */
    public static long parseLong(String strValue) {
	long value = 0;

	try {
	    if(strValue.startsWith("0x")) {
		strValue = strValue.substring(2);
		value = Long.parseLong(strValue, 16);
	    }
	    else
		value = Long.parseLong(strValue);
	}
	catch(NumberFormatException exp) {
	    errdebugln("Invalid number format: " + strValue);
	    throw exp;
	}

	return(value);
    }

    /**
     * Convert a string to a boolean value.
     * @param strValue The string
     * @return The value
     */
    public static boolean parseBoolean(String strValue) {
	if(strValue != null && "1".equals(strValue.trim()) || "true".equals(strValue.trim()) || "TRUE".equals(strValue.trim()))
	    return(true);
	return(false);
    }

    /**
     * Format integer to a string of specified width by prepending 
     * a suitable number of zeros.
     *
     * @param intValue The value
     * @param digits   The number of digits in the formatted string
     * @return The formatted string
     */
    public static String formatIntValue(int intValue, int digits) {
	String strValue = "";

	for(int i=digits; i>0; i--) {
	    if(Math.pow(10.0, (double) (i-1)) - intValue <= 0) {
		strValue = strValue + intValue;
		break;
	    }
	    strValue = strValue + "0";
	}

	return(strValue);
    }

    /**
     * Format hexadecimal byte to a 2 digit string by 
     * prepending a zero if needed.
     *
     * @param hexValue The value
     * @return The formatted string
     */
    public static String formatHexValue(byte hexValue) {
	if((hexValue & 0xff) < 16)
	    return("0" + Integer.toHexString(hexValue & 0xff));
	return(Integer.toHexString(hexValue & 0xff));
    }

    /**
     * Outputs a packetdump regardless of the debug level.
     * @param packet The packet to be dumped
     */
    public static void dumpPacket(Packet packet) {
	dumpPacket(0, packet.getBytes());
    }

    /**
     * Outputs a packetdump taking into account the debug level.
     * @param level  The debug level of this dump
     * @param packet The packet to be dumped
     */
    public static void dumpPacket(int level, Packet packet) {
	dumpPacket(level, packet.getBytes());
    }

    /**
     * Outputs a packetdump regardless of the debug level.
     * @param packet The bytes to be dumped
     */
    public static void dumpPacket(byte[] packetBuffer) {
	dumpPacket(0, packetBuffer);
    }

    /**
     * Outputs a packetdump taking into account the debug level.
     * @param level  The debug level of this dump
     * @param packet The bytes to be dumped
     */
    public  static void dumpPacket(int level, byte[] packetBuffer) {
	if(level <= debug) {
	    /* Output packet */
	    PacketUtils.debug(0, "\t");
	    for(int i=0; packetBuffer != null && i<packetBuffer.length; i++) {
		PacketUtils.debug(0, formatHexValue(packetBuffer[i]) + " ");
		if((i+1)%16==0)
		    PacketUtils.debug(0, "\n\t");
	    }
	    PacketUtils.debugln(0, "");
	}
    }

    /**
     * Construct a IPv4 pseudo header for checksum calculation.
     * @param ipv4Header The real header
     * @return An array of bytes containing the pseudo header
     */
    public static byte[] buildIPv4PseudoHeader(IPv4 ipv4Header, byte protocol, int segment_length) {
	byte[] pseudoHeader = new byte[12];
	byte[] ipv4Data = ipv4Header.getBytes();
	
	/* src */
	pseudoHeader[0] = ipv4Data[12];
	pseudoHeader[1] = ipv4Data[13];
	pseudoHeader[2] = ipv4Data[14];
	pseudoHeader[3] = ipv4Data[15];

	/* dst */
	pseudoHeader[4] = ipv4Data[16];
	pseudoHeader[5] = ipv4Data[17];
	pseudoHeader[6] = ipv4Data[18];
	pseudoHeader[7] = ipv4Data[19];

	/* padding */
	pseudoHeader[8] = 0;

	/* protocol */
	pseudoHeader[9] = protocol;

	/* segment length */
	pseudoHeader[10] = (byte) ((segment_length & 0xff00) >>> 8);
	pseudoHeader[11] = (byte) (segment_length & 0x00ff);
	
	PacketUtils.debugln(5, "PacketUtils.buildIPv4PseudoHeader()");

	return(pseudoHeader);
    }

    /**
     * Construct a IPv6 pseudo header for checksum calculation.
     * @param ipv6Header The real header
     * @return An array of bytes containing the pseudo header
     */
    public static byte[] buildIPv6PseudoHeader(IPv6 ipv6Header, Inet6Address src, Inet6Address dst, int payload_length, byte payload_proto) {
	byte[] pseudoHeader = new byte[40];
	byte[] ipv6Data = ipv6Header.getBytes();
	
	/* Source address */
	if(src != null) {
	    System.arraycopy(src.getAddress(), 0, pseudoHeader, 0, 16);
	}
	System.arraycopy(ipv6Data, 8, pseudoHeader, 0, 16);

	/* Destination address */
	if(dst != null) {
	    System.arraycopy(dst.getAddress(), 0, pseudoHeader, 16, 16);
	}
	else
	    System.arraycopy(ipv6Data, 24, pseudoHeader, 16, 16);

	/* Upper-Layer Packet Length */
	pseudoHeader[32] = (byte) ((payload_length & 0xff000000) >>> 24);
	pseudoHeader[33] = (byte) ((payload_length & 0x00ff0000) >>> 16);
	pseudoHeader[34] = (byte) ((payload_length & 0x0000ff00) >>> 8);
	pseudoHeader[35] = (byte) ((payload_length & 0x000000ff));
	
	/* Zero */
	pseudoHeader[36] = (byte) 0;
	pseudoHeader[37] = (byte) 0;
	pseudoHeader[38] = (byte) 0;

	/* Next Header */
	pseudoHeader[39] = payload_proto;

	PacketUtils.debugln(5, "PacketUtils.buildIPv6PseudoHeader()");

	return(pseudoHeader);
    }

    /**
     * Calculate UDP checksum.
     * @param pseudoHeader   The pseudoHeader
     * @param header         The UDP header 
     * @param payload        An bytearray containing the UDP payload data
     * @param payload_offset The startindex of payload
     * @param payload_length The length of the payload
     */
    public static int calcUDPChecksum(byte[] pseudoHeader, byte[] header, byte[] payload, int payload_offset, int payload_length) {
	return(calcTCPChecksum(pseudoHeader, header, payload, payload_offset, payload_length));
    }

    /**
     * Calculate TCP checksum.
     * @param pseudoHeader   The pseudoHeader
     * @param header         The TCP header 
     * @param payload        An bytearray containing the UDP payload data
     * @param payload_offset The startindex of payload
     * @param payload_length The length of the payload
     */
    public static int calcTCPChecksum(byte[] pseudoHeader, byte[] header, byte[] payload, int payload_offset, int payload_length) {
	if(payload_offset + payload_length > payload.length)
	    payload_length = payload.length - payload_offset;

	int checksum = 0;
	int i = 0;

	/* Sum pseudo/ICMP header */
	for(i=0; i<pseudoHeader.length-1; i+=2) {
	    checksum = addOnesComplement(checksum,  ((pseudoHeader[i] & 0xff) << 8) + (pseudoHeader[i+1] & 0xff));
	}
	if(i == pseudoHeader.length-1) {
	    checksum = addOnesComplement(checksum,  ((pseudoHeader[i] & 0xff) << 8) + (header[0] & 0xff));
	    i = 1;
	}
	else
	    i = 0;

	/* Sum TCP/UDP header */
	for(; i<header.length-1; i+=2) {
	    checksum = addOnesComplement(checksum,  ((header[i] & 0xff) << 8) + (header[i+1] & 0xff));
	}
	if(i == header.length-1) {
	    checksum = addOnesComplement(checksum,  ((header[i] & 0xff) << 8) + (payload[payload_offset] & 0xff));
	    i = payload_offset + 1;
	}
	else
	    i = payload_offset;

	/* Sum payload */
	for(; i<(payload_offset+payload_length-1); i+=2) {
	    checksum = addOnesComplement(checksum,  ((payload[i] & 0xff) << 8) + (payload[i+1] & 0xff));
	}
	if(i == (payload_offset+payload_length-1))
	    checksum = addOnesComplement(checksum,  ((payload[i] & 0xff) << 8));

	/* Take one's complement of the checksum */
	checksum = (checksum ^ 0xffff) & 0xffff;

	PacketUtils.debugln(3, "PacketUtils.calcChecksum() = 0x" + Integer.toHexString(checksum));

	return(checksum);
    }

    /**
     * Calculate IPv4 checksum.
     * @param header         The IPv4 header 
     * @param options        An bytearray containing the IPv4 options
     * @param options_offset The startindex of options
     * @param options_length The length of the options
     */
    public static int calcIPv4Checksum(byte[] header, byte[] options, int options_offset, int options_length) {
	if(options == null)
	    options_length = 0;
	else if(options_offset + options_length > options.length)
	    options_length = options.length - options_offset;

	int checksum = 0;
	int i = 0;

	/* Sum header */
	for(i=0; i<header.length-1; i+=2) {
	    checksum = addOnesComplement(checksum,  ((header[i] & 0xff) << 8) + (header[i+1] & 0xff));
	}
	if(i == header.length-1) {
	    checksum = addOnesComplement(checksum,  ((header[i] & 0xff) << 8) + (options[options_offset] & 0xff));
	    i = options_offset + 1;
	}
	else
	    i = options_offset;

	/* Sum options */
	for(; i<(options_offset+options_length-1); i+=2) {
	    checksum = addOnesComplement(checksum,  ((options[i] & 0xff) << 8) + (options[i+1] & 0xff));
	}
	if(i == (options_offset+options_length-1))
	    checksum = addOnesComplement(checksum,  ((options[i] & 0xff) << 8));
	    
	/* Take one's complement of the checksum */
	checksum = (checksum ^ 0xffff) & 0xffff;

	PacketUtils.debugln(3, "PacketUtils.calcIPv4Checksum() = 0x" + Integer.toHexString(checksum));

	return(checksum);
    }

    /**
     * Calculate ICMPv4 checksum.
     * @param header         The ICMPv4 header 
     * @param payload        An bytearray containing the rest of ICMP payload
     * @param payload_offset The startindex of options
     * @param payload_length The length of the options
     */
    public static int calcICMPv4Checksum(byte[] header, byte[] payload, int payload_offset, int payload_length) {

	if(payload == null) {
	    payload_offset = 0;
	    payload_length = 0;
	}

	int checksum = 0;
	int sum = 0;
	int i = 0;

	/* Sum header */
	for(i=0; i<header.length-1; i+=2) {
	    checksum = addOnesComplement(checksum,  ((header[i] & 0xff) << 8) + (header[i+1] & 0xff));
	}
	if(i == header.length-1) {
	    sum = (header[i] & 0xff) << 8;
	    i = 1;
	}
	else
	    i = 0;
    
	/* Sum payload */
	i += payload_offset;
	for(; i<(payload_offset+payload_length-1); i+=2) {
	    if(sum != 0) {
		checksum = addOnesComplement(checksum,  sum + (payload[payload_offset] & 0xff));
		sum = 0;
	    }
	    checksum = addOnesComplement(checksum,  ((payload[i] & 0xff) << 8) + (payload[i+1] & 0xff));
	}
	if(i == (payload_offset+payload_length-1))
	    checksum = addOnesComplement(checksum,  ((payload[i] & 0xff) << 8));
		
	if(sum != 0)
	    checksum = addOnesComplement(checksum, sum);
	
	/* Take one's complement of the checksum */
	checksum = (checksum ^ 0xffff) & 0xffff;

	PacketUtils.debugln(3, "PacketUtils.calcICMPv4Checksum() = 0x" + Integer.toHexString(checksum));

	return(checksum);
    }

    /**
     * Sum 16-bit integers a and b using one's complement arithmetic.
     * @param a Operand a
     * @param b Operand b
     * @return The result
     */
    private static int addOnesComplement(int a, int b) {
	int oa;
	int ob;
	int r;

	oa = (a & 0xffff);
	ob = (b & 0xffff);

	r = oa + ob;
	r = (r & 0xffff) + (r >> 16);
	r = ((r & 0xffff) + (r >> 16)) & 0xffff;

	return r;
    }

    /**
     * Set a unsigned byte in a given position of a given array 
     * to the given value
     *
     * @param buffer The array
     * @param index  Index of the u_int8
     * @param value  The value
     * @return false if an error occured, true otherwise
     */
    public static boolean setU8(byte[] buffer, int index, int value) {
	if(buffer == null || index < 0 || index+1 > buffer.length)
            return(false);
	buffer[index] = (byte) (value & 0xff);
	return(true);
    }

    /**
     * Set a unsigned 16-bit network order integer in a given 
     * positioen of a given array to the given value
     *
     * @param buffer The array
     * @param index  Index of the u_int16
     * @param value  The value
     * @return false if an error occured, true otherwise
     */
    public static boolean setU16(byte[] buffer, int index, int value) {
	if(buffer == null || index < 0 || index+2 > buffer.length)
            return(false);
	buffer[index] = (byte) ((value & 0xff00) >>> 8);
	buffer[index+1] = (byte) (value & 0x00ff);
	return(true);
    }

    /**
     * Set a unsigned 24-bit network order integer in a given 
     * positioen of a given array to the given value
     *
     * @param buffer The array
     * @param index  Index of the u_int16
     * @param value  The value
     * @return false if an error occured, true otherwise
     */
    public static boolean setU24(byte[] buffer, int index, int value) {
	if(buffer == null || index < 0 || index+3 > buffer.length)
            return(false);
	buffer[index] = (byte) ((value & 0xff0000) >>> 16);
	buffer[index+1] = (byte) ((value & 0x00ff00) >>> 8);
	buffer[index+2] = (byte) (value & 0x0000ff);
	return(true);
    }

    /**
     * Set a unsigned 32-bit network order integer in a given 
     * positioen of a given array to the given value
     *
     * @param buffer The array
     * @param index  Index of the u_int32
     * @param value  The value
     * @return false if an error occured, true otherwise
     */
    public static boolean setU32(byte[] buffer, int index, long value) {
	if(buffer == null || index < 0 || index+4 > buffer.length)
            return(false);
	buffer[index] = (byte)   ((value & 0xff000000) >>> 24);
	buffer[index+1] = (byte) ((value & 0x00ff0000) >>> 16);
	buffer[index+2] = (byte) ((value & 0x0000ff00) >>> 8);
	buffer[index+3] = (byte)  (value & 0x000000ff);
	return(true);
    }

    /**
     * Set a single bit.
     *
     * @param byteval The byte
     * @param index The index of the bit, 0 means least significant, 7 most significant
     * @return The modified byte
     */
    public static byte setBit(byte byteval, int index) {
	byte mask = ((byte) (1 << (index%8)));
	return(setMask(byteval, mask));
    }

    /**
     * Set a single bit.
     *
     * @param byteval The byte
     * @param index The index of the bit, 0 means least significant, 7 most significant
     * @return The modified byte
     */
    public static byte clearBit(byte byteval, int index) {
	byte mask = (byte) (0xff^((1 << (index%8)) & 0xff));
	return((byte) (byteval & mask));
    }
    
    /**
     * Set a single byte to a given bitmask.
     *
     * @param byteval The byte
     * @param mask The mask, 1 means set bit, 0 means leave bit alone
     * @return The modified byte
     */
    public static byte setMask(byte byteval, byte mask) {
	return((byte) (byteval | mask));
    }

    /**
     * Clear a single byte according to a given bitmask.
     *
     * @param byteval The byte
     * @param mask The mask, 0 means clear bit, 1 means leave bit alone
     * @return The modified byte
     */
    public static byte clearMask(byte byteval, byte mask) {
	return((byte) (byteval & mask));
    }

    /**
     * Extract a unsigned byte from an array of bytes.
     *
     * @param buffer The array
     * @param index  Index of the u_int8
     * @return The interger or -1 if an error occured
     */
    public static int getU8(byte[] buffer, int index) {
	if(buffer == null || index < 0 || index+1 > buffer.length)
            return(-1);
        return((int) (buffer[index] & 0xff));
    }

    /**
     * Extract a unsigned 16-bit network order integer from an array of bytes.
     *
     * @param buffer The array
     * @param index  Index of the u_int16
     * @return The interger or -1 if an error occured
     */
    public static int getU16(byte[] buffer, int index) {
	if(buffer == null || index < 0 || index+2 > buffer.length)
            return(-1);
        return((int) ((((buffer[index] & 0xff) << 8) | (buffer[index+1] & 0xff)) & 0x0000ffff));
    }

    /**
     * Extract an unsigned 24-bit network order integer from an array of bytes.
     *
     * @param buffer The array
     * @param index  Index of the u_int24
     * @return The interger or -1 if an error occured
     */
    public static int getU24(byte[] buffer, int index) {
	if(buffer == null || index < 0 || index+3 > buffer.length)
            return(-1);
        return((int) ((((buffer[index] & 0xff) << 16) | ((buffer[index] & 0xff) << 8) | (buffer[index+1] & 0xff)) & 0x00ffffff));
    }

    /**
     * Extract a unsigned 32-bit network order integer from an array of bytes.
     *
     * @param buffer The array
     * @param index  Index of the u_int32
     * @return The interger or -1 if an error occured
     */
    public static long getU32(byte[] buffer, int index) {
	if(buffer == null || index < 0 || index+4 > buffer.length)
            return(-1L);
	return((long) ((((buffer[index] & 0xff) << 24) | ((buffer[index+1] & 0xff) << 16) |
                        ((buffer[index+2] & 0xff) << 8) | (buffer[index+3] & 0xff))
                       & 0x00000000ffffffffL));
    }

    /**
     * Compare a single byte to a given bitmask.
     *
     * @param byteval The byte
     * @param index The index of the bit, 0 means least significant, 7 most significant
     * @return True if bit = 1, false otherwise
     */
    public static boolean getBit(byte byteval, int index) {
	byte mask = ((byte) (1 << (index%8)));
	return((byteval & mask) == mask);
    }

    /**
     * Compare a single byte to a given bitmask.
     *
     * @param byteval The byte
     * @param mask The mask
     * @return True if bit = 1, false otherwise
     */
    public static boolean getMask(byte byteval, byte mask) {
	return((byteval & mask) == mask);
    }
}
