/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine.header;

import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

public abstract class Header extends DefaultHandler {

    /* public variables */
    // none

    /* private variables */
    // none

    /* protected variables */
    protected byte[] buffer;
    protected int start;
    protected int end;

    protected DefaultHandler parent;
    protected SAXParser parser;

    protected String tag;
    
    /* constructor */

    public Header() {
	buffer = null;
	start = 0;
	end = 0;

        parent = null;
        parser = null;
        tag = null;
    }

    /* public methods */

    /**
     * Sets the XML parser handler.
     *
     * @param parent The parent handler
     * @param parser The parser
     */
    public final void setHandlers(DefaultHandler parent, SAXParser parser) {
        this.parent = parent;
        this.parser = parser;
    }

    /**
     * Returns the length of this header in bytes.
     *
     * @return The length
     */
    public final int getLength() {
        return(end-start);
    }

    /**
     * Returns the type tag of this header object.
     * Real protocols MUST implement this.
     *
     * @return The type tag
     */
    public abstract String getHeaderType();

    /**
     * Returns the human-readable name of this header class.
     * Real protocols MUST override this.
     *
     * @return The name
     */
    public static String getName() {
	return(NAME_NAME);
    }

    /**
     * Returns the type tag of this header class.
     * Real protocols MUST override this.
     *
     * @return The type tag
     */
    public static String getType() {
	return(TAG_NAME);
    }

    /**
     * Returns the tags and respective names that this header can parse.
     * Real protocols SHOULD override this, if applicable.
     *
     * @return A LinkedHashMap that contains the (tag, name) pairs.
     */
    public static LinkedHashMap getTags() {
        LinkedHashMap h = new LinkedHashMap();
        return(h);
    }

    /**
     * Returns the protocol id of this header.
     * Real protocols SHOULD override this, if applicable. 
     * By default returns UNDEFINED_PROTOCOL.
     * 
     * @return The id
     */
    public byte getProtocolID() {
        return(PROTOCOL_ID);
    }

    /**
     * Builds this header.
     * Real protocols SHOULD override this.
     *
     * @param buffer The buffer that contains all headers
     * @param offset Index of the successor header 
     * @param length Length of the buffer
     * @param headerList A List that contains all headers
     * @param index Index of this header
     *
     * @return The number of bytes added/removed due the rearrangement of headers
     */
    public int build(byte[] buffer, int offset, int length, LinkedList headerList, int index) {
	return(0);
    }

    /**
     * Returns an byte array containing the header.
     *
     * @return The array
     */
    public final byte[] getBytes() {
        byte[] packet = new byte[end-start];
        System.arraycopy(buffer, start, packet, 0, end-start);
        return(packet);
    }

    /**
     * Copies this header into the given byte array.
     *
     * @param dst The byte array
     * @param start_pos The index of the first byte
     * @param length Maximum length reserved for this header
     * 
     * @return The number of bytes copied, or -1 if something went wrong.
     */
    public final int copyBytes(byte[] dst, int start_pos, int length) {
        if(end-start > length)
            return -1;
	if(end-start <= 0)
	    return 0;

        System.arraycopy(buffer, start, dst, start_pos, end-start);
        return(end-start);
    }

    /* DefaultHandler methods */

    /**
     * This method is called during XML parsing, when a tag begins.
     * 
     * @param uri
     * @param localName
     * @param qName
     * @param attributes
     */
    public final void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
        tag = qName;
    }

    /**
     * This method is called during XML parsing, when the tags content is encountered.
     * Real protocols SHOULD override this.
     * 
     * @param ch The array of characters
     * @param offset The index of first character in the array
     * @param length Length of the content 
     */
    public void characters(char[] ch, int offset, int length) throws SAXException {
    }

    /**
     * This method is called during XML parsing, when a tag ends.
     * 
     * @param uri
     * @param localName
     * @param qName
     * @param attributes
     */
    public final void endElement(String uri, String localName, String qName) throws SAXException {
        if(Header.TAG_HEADER.equals(qName)) {

            try {
                parser.getXMLReader().setContentHandler(parent);
            }
            catch(Exception e) {
                System.err.println("FATAL:");
                e.printStackTrace();
                System.exit(-1);
            }
        }

        else
            tag = null;
    }

    /* private methods */
    // none

    /* static private variables */

    public static String TAG_NAME = "";
    public static String NAME_NAME = "";

    /* static public variables */

    public static byte PROTOCOL_ID = Header.UNDEFINED_PROTOCOL;

    public static final String TAG_DOC = "packetset";
    public static final String NAME_DOC = "packetset";

    public static final String TAG_PACKET = "packet";
    public static final String NAME_PACKET = "packet";

    public static final String TAG_PACKET_ATTR_STIME = "stime";

    public static final String TAG_HEADER = "header";
    public static final String NAME_HEADER = "header";

    public static final String TAG_HEADER_ATTR_TYPE = "type";

    public static final String PACKAGE_PREFIX = "packetengine.header";

    public static final byte NO_NEXTHEADER = (byte) 59;
    public static final byte UNDEFINED_PROTOCOL = (byte) 0;
}



