/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine.header;

import packetengine.*;
import java.util.*;
import java.net.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;


public class ICMPv4 extends Header {

    // ICMP specific stuff
    private byte[] ipv4Header;
    private byte[] payload;
    private int payload_offset;
    private int payload_length;


    public ICMPv4() {
	super();

        buffer = new byte[8];
        start = 0;
        end = 8;

        ipv4Header = null;
        payload = null;
        payload_offset = 0;
        payload_length = 0;
    }

    /* Header methods */

    public static String getName() {
        return(NAME_NAME);
    }

    public static String getType() {
        return(TAG_NAME);
    }

    public String getHeaderType() {
        return(TAG_NAME);
    }

    public byte getProtocolID() {
	return(PROTOCOL_ID);
    }
    
    public static LinkedHashMap getTags() {
        LinkedHashMap h = new LinkedHashMap();
	h.put(TAG_TYPE, NAME_TYPE);
	h.put(TAG_CODE, NAME_CODE);
	h.put(TAG_CSUM, NAME_CSUM);
	h.put(TAG_ECHOREQ, NAME_ECHOREQ);
	h.put(TAG_ECHOREP, NAME_ECHOREP);
        return(h);
    }

    public int build(byte[] buffer, int offset, int length, LinkedList headerList, int index) {
	if(buffer != null && offset >= 0 && offset+length <= buffer.length)	   
	    setPayload(buffer, offset, length);

	if(!FLAG_CSUM) {
	    setChecksum(0);
	    setChecksum(PacketUtils.calcICMPv4Checksum(getBytes(), payload, payload_offset, payload_length));
	    payload = null;
        }

	return(0);
    }

    /* DefaultHandler methods */

    public void characters(char[] ch, int offset, int length) throws SAXException {
        if(TAG_TYPE.equals(tag)) {
            PacketUtils.debugln(5, TAG_TYPE + ": " + new String(ch, offset, length));
            setType((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
        }
	else if(TAG_CODE.equals(tag)) {
	    PacketUtils.debugln(5, TAG_CODE+ ": " + new String(ch, offset, length));
            setCode((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
	}
	else if(TAG_CSUM.equals(tag)) {
	    PacketUtils.debugln(5, TAG_CSUM + ": " + new String(ch, offset, length));
	    FLAG_CSUM = false;
            setChecksum(PacketUtils.parseInt(new String(ch, offset, length)));
	    FLAG_CSUM = true;
	}
	else if(TAG_ECHOREQ.equals(tag)) {
	    PacketUtils.debugln(5, TAG_ECHOREQ+ ": " + new String(ch, offset, length));
            setEcho(new String(ch, offset, length), 8);
	}
	else if(TAG_ECHOREP.equals(tag)) {
	    PacketUtils.debugln(5, TAG_ECHOREP+ ": " + new String(ch, offset, length));
            setEcho(new String(ch, offset, length), 0);
	}
    }

    /* ICMP methods */

    public void setType(byte type) {
	buffer[start] = type;
    }

    public void setCode(byte code) {
	buffer[start+1] = code;
    }

    public void setChecksum(int csum) {
	if(!FLAG_CSUM) {
	    buffer[start+2] = (byte) ((csum & 0xff00) >>> 8);
	    buffer[start+3] = (byte) (csum & 0x00ff);	
	}
    }

    public void setIdentifier(int id) {
	    buffer[start+4] = (byte) ((id & 0xff00) >>> 8);
	    buffer[start+5] = (byte) (id & 0x00ff);	
    }

    public void setSequence(int seq) {
	    buffer[start+6] = (byte) ((seq & 0xff00) >>> 8);
	    buffer[start+7] = (byte) (seq & 0x00ff);	
    }

    public void setEcho(String params, int type) {
	setType((byte) (type & 0xff));
	setCode((byte) 0);
	
	StringTokenizer stok = new StringTokenizer(params, ",");
	while(stok.hasMoreTokens()) {
	    String token = (String) stok.nextToken(); 
	    if(!"".equals(token)) {

		int i = token.indexOf("=");
		if(i!=-1) {
		    String param = token.substring(0, i).trim();
		    String strValue = token.substring(i+1).trim();
		    int value = PacketUtils.parseInt(strValue);
		    if(TAG_ID.equals(param))
			setIdentifier(value);
		    if(TAG_SEQ.equals(param)) 
			setSequence(value);
		}
	    }
	}
    }

    public void setPayload(byte[] data, int offset, int length) {
        payload = data;
        payload_offset = offset;
        payload_length = length;
    }

    /* private ICMP methods */


    /* name */
    private static final String TAG_NAME = "ICMPv4";
    private static final String NAME_NAME = "Internet Control Message Protocol Version 4";

    /* tags */
    private static final String TAG_TYPE = "type";
    private static final String NAME_TYPE = "Type (0..256)";

    private static final String TAG_CODE = "code";
    private static final String NAME_CODE = "Code (0..256)";

    private static final String TAG_CSUM = "csum";
    private static final String NAME_CSUM = "Checksum (0x0...0xffff)";
    private boolean FLAG_CSUM = false;

    private static final String TAG_ECHOREQ = "echoreq";
    private static final String NAME_ECHOREQ = "Echo Request (id, seq)";

    private static final String TAG_ECHOREP = "echorep";
    private static final String NAME_ECHOREP = "Echo Reply (id, seq)";

    private static final String TAG_ID = "id";
    private static final String TAG_SEQ = "seq";

    /* protocol id */
    public static final byte PROTOCOL_ID = (byte) 1;
}

