/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine.header;

import packetengine.*;
import java.util.*;
import java.net.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

public class IPv4 extends Header {

    /* private variables */
    private byte[] options;
    private int options_offset;
    private int options_length;

    public IPv4() {
	super();

	/* create buffer */
	buffer = new byte[20];
	start = 0;
	end = 20;
	setVersion();

	options = null;
	options_offset = 0;
	options_length = 0;

	/* Set defaults */
	setVersion();
	setIdentification((new Random()).nextInt());
	setTTL((byte) 1);
	try {
	    setSourceAddress(Inet4Address.getByName("127.0.0.1").getAddress());
	    setDestinationAddress(Inet4Address.getByName("127.0.0.1").getAddress());
	}
	catch(UnknownHostException e) {
	    PacketUtils.errdebugln(e.toString());
	}
    }

    public IPv4(byte[] buffer, int offset) {
	super();

	/* set buffer */
	this.buffer = buffer;
	start = offset;
	end = start + 20;

	options = null;
	options_offset = 0;
	options_length = 0;
    }

    /* Header methods */

    public static String getName() {
        return(NAME_NAME);
    }

    public static String getType() {
        return(TAG_NAME);
    }

    public String getHeaderType() {
        return(TAG_NAME);
    }

    public byte getProtocolID() {
	return(PROTOCOL_ID);
    }

    public static LinkedHashMap getTags() {
	LinkedHashMap h = new LinkedHashMap();
	h.put(TAG_VERSION, NAME_VERSION);
	h.put(TAG_IHL, NAME_IHL);
	h.put(TAG_TOS, NAME_TOS);
	h.put(TAG_LENGTH, NAME_LENGTH);
	h.put(TAG_ID, NAME_ID);
	h.put(TAG_DF, NAME_DF);
	h.put(TAG_MF, NAME_MF);
	h.put(TAG_FRAGOFFSET, NAME_FRAGOFFSET);
	h.put(TAG_TTL, NAME_TTL);
	h.put(TAG_PROTO, NAME_PROTO);
	h.put(TAG_CSUM, NAME_CSUM);
	h.put(TAG_SADDR, NAME_SADDR);
	h.put(TAG_DADDR, NAME_DADDR);
	return(h);
    }

    public int build(byte[] buffer, int offset, int length, LinkedList headerList, int index) {
	Header h = null;
	int opt_len, i;

	/* Handle IPv4 options */

	/* Count the length of option headers */
	opt_len = 0;
	for(i=index+1; i<headerList.size(); i++) {
	    h = (Header) headerList.get(i);
	    if(h != null && h instanceof IPv4Option)
		opt_len += h.getLength();
	    else
		break;
	}
	int ptr = offset+opt_len; // Pointer to first non-IPv4 option header

	/* If necessary add EOL options */
	int fixed_opt_len = ((int) Math.ceil( (double) opt_len / (double) 4 ) ) * 4;
	int eol_count = 0;
	while(opt_len < fixed_opt_len) {
	    IPv4_OPT_EOL eol = new IPv4_OPT_EOL();
	    headerList.add(i, eol);
	    eol.build(buffer, ptr, length-ptr, headerList, i);
	    opt_len += eol.getLength();
	    ptr -= eol.copyBytes(buffer, ptr-eol.getLength(), eol.getLength());
	    eol_count++;
	}

	PacketUtils.debugln(5, "IPv4 - eol_count = " + eol_count);

	/* Rebuild options if they were moved by extra EOLs */
	if(eol_count > 0) {
	    for(int j=i-1; j>index; j--) {
		h = (Header) headerList.get(j);
		h.build(buffer, ptr, length-ptr, headerList, j);
		ptr -= h.copyBytes(buffer, ptr-h.getLength(), h.getLength());
	    }
	}
	else
	    ptr = offset;

	if(opt_len > 0)
	    setOptions(buffer, ptr, opt_len);

	/* Handle IPv4 */

	/* Set protocol field */
	try {
	    h = (Header) headerList.get(i+eol_count);
	    if(!FLAG_PROTO && h != null && h.getProtocolID() != Header.UNDEFINED_PROTOCOL)
		setProtocol(h.getProtocolID());
	}
	catch(IndexOutOfBoundsException e) {
	    /* No next header, leave empty? */
	}

	PacketUtils.debugln(5, "length      = " + length);
	PacketUtils.debugln(5, "ptr         = " + ptr);
	PacketUtils.debugln(5, "offset      = " + offset);
	PacketUtils.debugln(5, "getLength() = " + getLength());	

	/* Set other defaults */
	if(!FLAG_IHL)
	    setIHL((byte) ((int) (Math.ceil( (double) (getLength() + options_length) / (double) 4)) & 0xff));
	if(!FLAG_LENGTH)
	    setTotalLength(length + (offset - ptr) + getLength());
	if(!FLAG_CSUM) {
	    setChecksum(0);
	    setChecksum(PacketUtils.calcIPv4Checksum(getBytes(), options, options_offset, options_length));
	}

	return(offset - ptr);
    }

    /* DefaultHandler methods */

    public void characters(char[] ch, int offset, int length) throws SAXException {
	if(TAG_VERSION.equals(tag)) {
	    PacketUtils.debugln(5, TAG_VERSION + ": " + new String(ch, offset, length));
	    FLAG_VERSION = false;
	    setVersion((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
	    FLAG_VERSION = true;
	}
	else if(TAG_IHL.equals(tag)) {
	    PacketUtils.debugln(5, TAG_IHL + ": " + new String(ch, offset, length));
	    FLAG_IHL = false;
	    setIHL((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
	    FLAG_IHL = true;
	}
	else if(TAG_TOS.equals(tag)) {
	    PacketUtils.debugln(5, TAG_TOS + ": " + new String(ch, offset, length));
	    setTOS((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
	}
	else if(TAG_LENGTH.equals(tag)) {
	    PacketUtils.debugln(5, TAG_LENGTH + ": " + new String(ch, offset, length));
	    FLAG_LENGTH = false;
	    setTotalLength(PacketUtils.parseInt(new String(ch, offset, length)));
	    FLAG_LENGTH = true;
	}
	else if(TAG_ID.equals(tag)) {
	    PacketUtils.debugln(5, TAG_ID + ": " + new String(ch, offset, length));
	    FLAG_ID = false;
	    setIdentification(PacketUtils.parseInt(new String(ch, offset, length)));
	    FLAG_ID = true;
	}
	else if(TAG_DF.equals(tag)) {
	    PacketUtils.debugln(5, TAG_DF + ": " + new String(ch, offset, length));
	    if(PacketUtils.parseBoolean(new String(ch, offset, length)))
		setDF(true);
	    else
		setDF(false);
	}
	else if(TAG_MF.equals(tag)) {
	    PacketUtils.debugln(5, TAG_MF + ": " + new String(ch, offset, length));
	    if(PacketUtils.parseBoolean(new String(ch, offset, length)))
		setMF(true);
	    else
		setMF(false);
	}
	else if(TAG_FRAGOFFSET.equals(tag)) {
	    PacketUtils.debugln(5, TAG_FRAGOFFSET + ": " + new String(ch, offset, length));
	    setFragmentOffset(PacketUtils.parseInt(new String(ch, offset, length)));	    
	}
	else if(TAG_TTL.equals(tag)) {
	    PacketUtils.debugln(5, TAG_TTL + ": " + new String(ch, offset, length));
	    FLAG_TTL = false;
	    setTTL((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
	    FLAG_TTL = true;
	}
	else if(TAG_PROTO.equals(tag)) {
	    PacketUtils.debugln(5, TAG_PROTO + ": " + new String(ch, offset, length));
	    FLAG_PROTO = false;
	    setProtocol((byte)(PacketUtils.parseInt(new String(ch, offset, length)) & 0xff));
	    FLAG_PROTO = true;
	}
	else if(TAG_CSUM.equals(tag)) {
	    PacketUtils.debugln(5, TAG_CSUM + ": " + new String(ch, offset, length));
	    FLAG_CSUM = false;
	    setChecksum(PacketUtils.parseInt(new String(ch, offset, length)));
	    FLAG_CSUM = true;
	}
	else if(TAG_SADDR.equals(tag)) {
	    PacketUtils.debugln(5, TAG_SADDR + ": " + new String(ch, offset, length));
	    FLAG_SADDR = false;
	    try {
		setSourceAddress(Inet4Address.getByName(new String(ch, offset, length)).getAddress());
	    }
	    catch(UnknownHostException e) {
		PacketUtils.errdebugln(e.toString());
		throw new SAXException(e);
	    }
	    FLAG_SADDR = true;
	}
	else if(TAG_DADDR.equals(tag)) {
	    PacketUtils.debugln(5, TAG_DADDR + ": " + new String(ch, offset, length));
	    FLAG_DADDR = false;
	    try {
		setDestinationAddress(Inet4Address.getByName(new String(ch, offset, length)).getAddress());
	    }
	    catch(UnknownHostException e) {
		PacketUtils.errdebugln(e.toString());
		throw new SAXException(e);
	    }
	    FLAG_DADDR = true;
	}
    }

    /* IPv4 methods */

    public void setVersion() {
        buffer[start] = (byte) ( 0x40 | (buffer[start] & 0x0f));
    }

    public void setVersion(byte version) {
	if(!FLAG_VERSION)
	    buffer[start] = (byte) (( version & 0x0f ) << 4 | (buffer[start] & 0x0f));
    }

    public void setIHL(byte length) {
	if(!FLAG_IHL)
	    buffer[start] = (byte) (( length & 0x0f ) | (buffer[start] & 0xf0));
    }

    public void setTOS(byte tos) {
	PacketUtils.setU8(buffer, start+1, tos);
    }

    public void setTotalLength(int length) {
	if(!FLAG_LENGTH)
	    PacketUtils.setU16(buffer, start+2, length);
    }
    
    public void setIdentification(int id) {
	if(!FLAG_ID)
	    PacketUtils.setU16(buffer, start+4, id);
    }
    
    public void setDF(boolean df) {
	if(df)
	    buffer[start+6] = PacketUtils.setMask(buffer[start+6], (byte) 0x40);
	else
	    buffer[start+6] = PacketUtils.setMask(buffer[start+6], (byte) 0xbf);
    }

    public void setMF(boolean mf) {
	if(mf)
	    buffer[start+6] = PacketUtils.setMask(buffer[start+6], (byte) 0x20);
	else
	    buffer[start+6] = PacketUtils.setMask(buffer[start+6], (byte) 0xdf);
    }

    public void setFragmentOffset(int offset) {
	buffer[start+6] = (byte) ((buffer[start+6] & 0xe0) |
				  (((offset & 0xff00) >>> 8) & 0x1f));
	buffer[start+7] = (byte) (offset & 0x00ff);
    }

    public void setTTL(byte ttl) {
	if(!FLAG_TTL)
	    PacketUtils.setU8(buffer, start+8, ttl);
    }

    public void setProtocol(byte proto) {
	if(!FLAG_PROTO)
	    PacketUtils.setU8(buffer, start+9, proto);
    }

    public void setChecksum(int csum) {
	if(!FLAG_CSUM)
	    PacketUtils.setU16(buffer, start+10, csum);
    }

    public void setSourceAddress(Inet4Address address) {
	if(address != null)
	    setSourceAddress(address.getAddress());
    }

    public void setSourceAddress(byte[] address) {
	if(FLAG_SADDR || address.length < 4)
            return;

        for(int i=0; i<4; i++)
            buffer[start+12+i] = address[i];
    }

    public void setDestinationAddress(Inet4Address address) {
	if(address != null)
	    setDestinationAddress(address.getAddress());
    }

    public void setDestinationAddress(byte[] address) {
	if(FLAG_DADDR || address.length < 4)
            return;

        for(int i=0; i<4; i++)
            buffer[start+16+i] = address[i];
    }
    
    public void setOptions(byte[] data, int offset, int length) {
	options = data;
	options_offset = offset;
	options_length = length;
    }

    public byte getIHL() {
        return((byte) (( buffer[start] & 0x0f )));
    }

    public int getTotalLength() {
	return((int) ((buffer[start+2] << 8) | buffer[start+3]));
    }

    public byte getTOS() {
	return getTOS(buffer, start, end-start);
    }

    public byte getTTL() {
	return getTTL(buffer, start, end-start);
    }

    public byte getProtocol() {
	return getProtocol(buffer, start, end-start);
    }

    public Inet4Address getSourceAddress() {
	return getSourceAddress(buffer, start, end-start);
    }

    public Inet4Address getDestinationAddress() {
	return getDestinationAddress(buffer, start, end-start);
    }

    /* static public methods */

    public static boolean identify(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(false);
	if((data[offset] & 0xf0) == 0x40)
	    return(true);
	return(false);
    }

    public static byte getIHL(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(-1);
	return((byte)(data[offset] & 0x0f));
    }

    public static byte getTOS(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(-1);
	return(data[offset+1]);
    }

    public static int getTotalLength(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(-1);
	return(PacketUtils.getU16(data, offset+2));
    }

    public static byte getTTL(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(-1);
	return(data[offset+8]);
    }

    public static byte getProtocol(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(-1);
	return(data[offset+9]);
    }

    public static Inet4Address getSourceAddress(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(null);
	byte[] addr = new byte[4];
	System.arraycopy(data, offset+12, addr, 0, 4);
	try {
	    return((Inet4Address) Inet4Address.getByAddress(addr));
	}
	catch(UnknownHostException exp) {
	}
	return(null);
    }

    public static Inet4Address getDestinationAddress(byte[] data, int offset, int length) {
	if(data == null || offset < 0 || length < 20)
	    return(null);
	byte[] addr = new byte[4];
	System.arraycopy(data, offset+16, addr, 0, 4);
	try {
	    return((Inet4Address) Inet4Address.getByAddress(addr));
	}
	catch(UnknownHostException exp) {
	}
	return(null);
    }
    
    /* private methods */

    /* variables */

    /* name */
    public static final String TAG_NAME = "IPv4";
    public static final String NAME_NAME = "IPv4";

    /* tags */
    private static final String TAG_VERSION = "version";
    private static final String NAME_VERSION = "Version (0...15)";
    private boolean FLAG_VERSION = false;

    private static final String TAG_IHL = "ihl";
    private static final String NAME_IHL = "Header Length (0...15)";
    private boolean FLAG_IHL = false;

    private static final String TAG_TOS = "tos";
    private static final String NAME_TOS = "Type of Service (0x00...0xff)";

    private static final String TAG_LENGTH = "length";
    private static final String NAME_LENGTH = "Total Length (0...65535)";
    private boolean FLAG_LENGTH = false;

    private static final String TAG_ID = "id";
    private static final String NAME_ID = "Identification (0...65535)";
    private boolean FLAG_ID = false;

    private static final String TAG_DF = "df";
    private static final String NAME_DF = "Don't Fragment (true/false)";

    private static final String TAG_MF = "mf";
    private static final String NAME_MF = "More Fragments (true/false)";

    private static final String TAG_FRAGOFFSET = "fragoffset";
    private static final String NAME_FRAGOFFSET = "Fragment Offset (0...8191)";

    private static final String TAG_TTL = "ttl";
    private static final String NAME_TTL = "Time to Live (0...255)";
    private boolean FLAG_TTL = false;

    private static final String TAG_PROTO = "proto";
    private static final String NAME_PROTO = "Protocol (0...255)";
    private boolean FLAG_PROTO = false;
    
    private static final String TAG_CSUM = "csum";
    private static final String NAME_CSUM = "Checksum (0x0...0xffff)";
    private boolean FLAG_CSUM = false;

    private static final String TAG_SADDR = "srcaddr";
    private static final String NAME_SADDR = "Source Address (hostname/IP)";
    private boolean FLAG_SADDR = false;

    private static final String TAG_DADDR = "dstaddr";
    private static final String NAME_DADDR = "Destination Address (hostname/IP)";
    private boolean FLAG_DADDR = false;

    /* protocol id */
    public static final byte PROTOCOL_ID = (byte) 4;
}

