/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or 
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified 
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package packetengine.header;

import packetengine.*;
import java.util.*;
import java.net.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

public class UDP extends Header {

    // UDP specific stuff
    private byte[] pseudoHeader;
    private byte[] payload;
    private int payload_offset;
    private int payload_length;

    public UDP() {
	super();
	
	/* create buffer */
        buffer = new byte[8];
        start = 0;
        end = 8;

	pseudoHeader = null;
	payload = null;
	payload_offset = 0;
	payload_length = 0;
    }

    public UDP(byte[] buffer, int offset) {
	super();
	
	/* create buffer */
        this.buffer = buffer;
        start = offset;
        end = start + 8;

	pseudoHeader = null;
	payload = null;
	payload_offset = 0;
	payload_length = 0;
    }


    /* Header methods */

    public static String getName() {
        return(NAME_NAME);
    }

    public static String getType() {
        return(TAG_NAME);
    }

    public String getHeaderType() {
        return(TAG_NAME);
    }

    public byte getProtocolID() {
	return(PROTOCOL_ID);
    }
    
    public static LinkedHashMap getTags() {
        LinkedHashMap h = new LinkedHashMap();
        h.put(TAG_SPORT, NAME_SPORT);
        h.put(TAG_DPORT, NAME_DPORT);
        h.put(TAG_LENGTH, NAME_LENGTH);
        h.put(TAG_CSUM, NAME_CSUM);
        return(h);
    }
    
    public int build(byte[] buffer, int offset, int length, LinkedList headerList, int index) {
	Header h = null;

	if(buffer != null && offset >= 0 && offset+length <= buffer.length)
	    setPayload(buffer, offset, length);

	/* csum UDP */
	if(headerList != null && index < headerList.size()) {
	    for(int j=index-1; j>=0; j--) {
		h = (Header) headerList.get(j);
		if(h instanceof IPv4 || h instanceof IPv6)
		    break;
		h = null;
	    }
	    if(h != null && h instanceof IPv4)
		setPseudoHeader(PacketUtils.buildIPv4PseudoHeader((IPv4) h, PROTOCOL_ID, payload_length + getLength()));
	    else if(h != null && h instanceof IPv6) {
		int payload_length = getPayloadLength();
		setPseudoHeader(PacketUtils.buildIPv6PseudoHeader((IPv6) h, null, null, payload_length, PROTOCOL_ID)); 
	    }
	}
	
	if(!FLAG_LENGTH)
	    setLength(payload_length + getLength());

	if(!FLAG_CSUM) {
	    if(payload != null && pseudoHeader != null) {
		setChecksum(0);
		setChecksum(PacketUtils.calcUDPChecksum(pseudoHeader, getBytes(), payload, payload_offset, payload_length));
		pseudoHeader = null;
		payload = null;
	    }
	}

	return(0);
    }

    /* DefaultHandler methods */

    public void characters(char[] ch, int offset, int length) throws SAXException {
	if(TAG_SPORT.equals(tag)) {
            PacketUtils.debugln(5, TAG_SPORT + ": " + new String(ch, offset, length));
            setSourcePort(PacketUtils.parseInt(new String(ch, offset, length)));
        }
        else if(TAG_DPORT.equals(tag)) {
            PacketUtils.debugln(5, TAG_DPORT + ": " + new String(ch, offset, length));
            setDestinationPort(PacketUtils.parseInt(new String(ch, offset, length)));
        }
        else if(TAG_LENGTH.equals(tag)) {
            PacketUtils.debugln(5, TAG_LENGTH + ": " + new String(ch, offset, length));
	    FLAG_LENGTH = false;
            setLength(PacketUtils.parseInt(new String(ch, offset, length)));
	    FLAG_LENGTH = true;
        }
        else if(TAG_CSUM.equals(tag)) {
            PacketUtils.debugln(5, TAG_CSUM + ": " + new String(ch, offset, length));
	    FLAG_CSUM = false;
            setChecksum(PacketUtils.parseInt(new String(ch, offset, length)));
	    FLAG_CSUM = true;
        }
    }

    /* UDP methods */

    public void setSourcePort(int port) {
	PacketUtils.setU16(buffer, start, port);
    }

    public void setDestinationPort(int port) {
	PacketUtils.setU16(buffer, start+2, port);
    }

    public void setLength(int length) {
	if(!FLAG_LENGTH)
	    PacketUtils.setU16(buffer, start+4, length);
    }

    public void setChecksum(int csum) {
	if(!FLAG_CSUM)
	    PacketUtils.setU16(buffer, start+6, csum);
    }

    public void setPseudoHeader(byte[] header) {
        pseudoHeader = header;
    }

    public void setPayload(byte[] data, int offset, int length) {
        payload = data;
	payload_offset = offset;
	payload_length = length;
    }

    public int getPayloadLength() {
	return(payload_length);
    }

    public int getSourcePort() {
	return getSourcePort(buffer, start, end-start);
    }

    public int getDestinationPort() {
	return getDestinationPort(buffer, start, end-start);
    }

    public int getUDPLength() {
	return getUDPLength(buffer, start, end-start);
    }

    /* static public methods */

    public static boolean identify(byte[] data, int offset, int length) {
        if(data == null || offset < 0 || length < 8)
            return(false);
        return(true);
    }

    public static int getSourcePort(byte[] data, int offset, int length) {
        if(data == null || offset < 0 || length < 8)
            return(-1);
	return(PacketUtils.getU16(data, offset));
    }

    public static int getDestinationPort(byte[] data, int offset, int length) {
        if(data == null || offset < 0 || length < 8)
            return(-1);
	return(PacketUtils.getU16(data, offset+2));
    }

    public static int getUDPLength(byte[] data, int offset, int length) {
        if(data == null || offset < 0 || length < 8)
            return(-1);
	return(PacketUtils.getU16(data, offset+4));
    }

    /* private UDP methods */

    /* name */
    private static final String TAG_NAME = "UDP";
    private static final String NAME_NAME = "Unreliable Datagram Protocol";

    /* tags */
    private static final String TAG_SPORT = "srcport";
    private static final String NAME_SPORT = "Source Port (0...65535)";

    private static final String TAG_DPORT = "dstport";
    private static final String NAME_DPORT = "Destination Port (0...65535)";

    private static final String TAG_LENGTH = "length";
    private static final String NAME_LENGTH = "Total Length (0...65535)";
    private boolean FLAG_LENGTH = false;

    private static final String TAG_CSUM = "csum";
    private static final String NAME_CSUM = "Checksum (0x0...0xffff)";
    private boolean FLAG_CSUM = false;

    /* protocol id */
    public static final byte PROTOCOL_ID = (byte) 17;
}
