/**
 *
 * Copyright (C) 2002-2003 VTT Information Technology. All rights reserved.
 *
 * This software is provided AS-IS with no warranty, either expressed or
 * implied.
 *
 * This software is distributed under licence and may not be copied, modified
 * or distributed except as expressly authorised under the terms of the licence
 * contained in the file LICENSE in this distribution.
 *
 * For more information about licensing, please refer to
 * http://opensource.erve.vtt.fi/ or contact VTT Information Technology.
 *
 */
package tcpstreamer;

import packetengine.*;
import java.net.Socket;
import java.net.InetSocketAddress;
import java.io.OutputStream;
import java.util.StringTokenizer;


/**
 * TCPStreamer is a network interface that creates a real TCP connection
 * to the remote host. The interface uses a normal Socket, thus IP and TCP
 * headers are created by the network stack.
 *
 * TCPStreamer accepts following parameters in the network interface dialog:
 * host - Remote host name or IP address
 * port - Remote port
 * delay - Delay in milliseconds before the socket is closed
 *
 * The parameter 'delay' is handy in case you want to wait for data from the
 * remote host. 
 */
public class TCPStreamer extends NetworkInterface {

    /**
     * Default constructor does nothing
     */
    public TCPStreamer() {
	tcpSocket = null;
	destAddr = null;
	closeDelay = 0;
    }

    // NetworkInterface methods

    public static String getIFTag() {
	return(TAG_IF);
    }

    public boolean setParams(String params) {
	if(params != null && params.length() > 0) {
	    try {
		String host = "";
		int port = -1;

		StringTokenizer tok = new StringTokenizer(params);
		while(tok.hasMoreTokens()) {
		    String str = tok.nextToken();
		    
		    if(str == null)
			;
		    else if(str.startsWith("host="))
			host = str.substring("host=".length());
		    else if(str.startsWith("port="))
			port = Integer.parseInt(str.substring("port=".length()));
		    else if(str.startsWith("delay="))
			closeDelay = Integer.parseInt(str.substring("delay=".length()));
		}

		destAddr = new InetSocketAddress(host, port);
		PacketUtils.debugln(5, "TCPStreamer: host=" + host 
				    + " port=" + port
				    + " delay=" + closeDelay);
		return(true);
	    }
	    catch(Exception exp) {
		PacketUtils.errdebugln("TCPStreamer parameter error: " + 
				       exp.toString());
	    }
	}
	return(false);
    }

    public boolean ifup() {
	try {
	    if(tcpSocket != null &&
	       tcpSocket.isConnected())
		tcpSocket.close();
	    
	    tcpSocket = new Socket();
	}
	catch(Exception exp) {
		PacketUtils.errdebugln("TCPStreamer error: " + 
				       exp.toString());
		return(false);
	}

	if(tcpSocket != null && destAddr != null) {
	    try {
		tcpSocket.connect(destAddr);
		PacketUtils.debugln(5, "TCPStreamer: Connection to "
				    + destAddr + " established.");
		return(true);
	    }
	    catch(Exception exp) {
		PacketUtils.errdebugln("TCPStreamer error: " + 
				       exp.toString());
	    }
	    return(false);
	}
	PacketUtils.errdebugln("TCPStreamer error: Socket not created.");
	return(false);
    }

    public boolean ifdown() {
	if(tcpSocket != null) {
	    if(closeDelay != 0) {
		try {
		    Thread.sleep(closeDelay);
		}
		catch(InterruptedException exp) {
		}
	    }

	    if(tcpSocket.isClosed()) {
		    PacketUtils.debugln(5, "TCPStreamer: Connection already closed.");
		    tcpSocket = null;
		    return(true);
	    }
	    else {		
		try {
		    tcpSocket.close();
		    PacketUtils.debugln(5, "TCPStreamer: Connection to " +
					destAddr + " closed.");
		    tcpSocket = null;
		    return(true);
		}
		catch(Exception exp) {
		    PacketUtils.errdebugln("TCPStreamer error: " + 
					   exp.toString());
		}
		return(false);
	    }
	}
	return(true);
    }

    public boolean sendData(byte[] packet) {
	if(tcpSocket != null && 
	   tcpSocket.isConnected() &&
	   packet != null) {
	    try {
		tcpSocket.getOutputStream().write(packet);
		PacketUtils.debugln(5, "TCPStreamer: " + packet.length +
				    " bytes sent.");
		return(true);
	    }
	    catch(Exception exp) {
		PacketUtils.errdebugln("TCPStreamer error: " + 
				       exp.toString());
	    }
	    return(false);
	}
	PacketUtils.errdebugln("TCPStreamer error: Packet not sent.");
	return(false);
    }

    // private PacketSender methods

    // private PacketSender variables
    private Socket tcpSocket;
    private InetSocketAddress destAddr;
    private int closeDelay;
 
    // static variables

    private static final String TAG_IF = "TCPStreamer";
}
